#IMPORTANT INFORMATION BEFORE STARTING

#.bam and .bai files should have the same name and should have identical paths as this .R file

#This script exports 4 different results in pzfx (Graphpad/Prism), xlsx (Excel) and csv files (lines 170-177, lines 314-315)
#You can choose 1 preferred file type by adding '#' before the other lines.

#PART 1: ANALYSIS PER LOCUS, exports the following files:
#file named as "-frequency-of-sequences" shows frequency of all sequences present in library
#file named as "-coverage-of-library-sequences" shows frequency of only the designed library variants
#file named as "-editing-efficiency" shows frequency of wild-type and library variants in general (all library variants merged into 1 class, 'library')

#PART 1: ANALYSIS PER LOCUS, exports the following file:
#file named as "-multiplex" shows number of mutated target loci per plasmid

#It is good practice to check how many reads were used for analysis.
#You can check this by looking at the 2 last printed objects in the console:
#Number of reads used for PART 1: ANALYSIS PER LOCUS --> print(numberofseq)
#Number of reads used for PART 2: ANALYSIS PER PLASMID --> print(nrow(df_select))

#Contact address for questions/problems during running the script: m.vandenbrink@tudelft.nl


#LET'S START!
#Install required packages: BiocManager, GenomicAlignments, ggplot2, pzfx, readxl, openxlsx
#install.packages("ggplot2")
#install.packages("readxl")
#install.packages("BiocManager")
#BiocManager::install("GenomicAlignments")
#install.packages("pzfx")
#install.packages("openxlsx")

#Load required packages
library(ggplot2)
library(readxl)
library(GenomicAlignments)
library(pzfx)
library(openxlsx)

#Clear environment
rm(list = ls())

#Essential information to be defined by user
folder = "/.../" #path to folder containing .bam, .bai and .R files
filename = "Galaxy238-13-07-A2" #name of the .bam/.bai files
indexfilename = "G555-pUC57-OriL191-lipid" #name of the reference file used for mapping in minimap2, can be checked in Galaxy by viewing minimap2 result
region = c("3249", #first nucleotide positions of target regions; vector should contain as many elements as target regions.
           "5934",
           "6881",
           "7977")
region_length = c(20, #length (in nucleotides) of each target region
                  20,
                  20,
                  20)
wildtype = c("TTTAAGAAGGAGATATACAT", #wildtype sequence of each target region
             "TTTAAGAAGGAGATATACAT",
             "TTTAAGAAGGAGATATACAT",
             "TTTAAGAAGGAGATATACAT")
predictedlibrary = c("PlsB-predicted-library-RBScalculator.xlsx", #.xlsx file containing one column with designed library variants per region
                     "PlsC-predicted-library-RBScalculator.xlsx",
                     "CdsA-PssA-predicted-library-RBScalculator.xlsx",
                     "CdsA-PssA-predicted-library-RBScalculator.xlsx")
qualityscore = 50 #minimum per-base quality threshold





### PART 1: ANALYSIS PER LOCUS

#Import .bam and .bai files
bamfile = paste(folder, filename, ".bam", sep = "")
indexfile = paste(folder, filename, sep = "") #indexfile is the .bai file

#Extract all sequences per target region
mylist1 <- list()
mylist2 <- list()
efficiency <- c()
wildtypelist <- c()
numberofseq <- c()
for (i in 1:length(region)){
  range = paste(indexfilename, ":", region[i], "-", as.character(as.numeric(region[i])+region_length[i]-1), sep = "") #define region
  stack <- stackStringsFromBam(bamfile, indexfile, param=GRanges(range)) # Extract the stacked reads from the position
  stack_seq = as.character(stack)
  
  qual <- stackStringsFromBam(bamfile, indexfile, param=GRanges(range), what="qual") # Extract the stacked reads from the position
  qual_seq = as.character(qual)
  
  #Quality control: eliminate sequences with at least 1 nucleotide with a per-base quality score lower than the threshold
  for (j in (1:length(qual_seq))) {
    if (min(strtoi(charToRaw(qual_seq[length(qual_seq)+1-j]),16L)) < qualityscore) {
      stack_seq = stack_seq[-(length(qual_seq)+1-j)]
    }
  }
  numberofseq[i] <- length(qual_seq) #save number of reads used for analysis for later

  #Make a list of all DNA variants in the target region
  df = c()
  for (k in 1:length(stack_seq)) {
    if (!(stack_seq[k] %in% df)) {df <- rbind(df, c(stack_seq[k], 1))}
    else {
      column = which(grepl(stack_seq[k], df))
      df[column, 2] <- as.numeric(df[column, 2])+1}
  }
  
  #Transform (to make compatible with column graph in Prism)
  df <- as.data.frame(df)
  df[,2]<-as.numeric(as.character(df[,2]))
  df <- df[order(df[,2], decreasing = TRUE),]
  df_frequency <- as.data.frame(t(df[,2]))
  colnames(df_frequency) <- df[,1]
  
  #Compare sequences with designed library sequences
  filename_predictedlibrary <- paste(folder, predictedlibrary[i], sep = "")
  predictedlib <- read_excel(filename_predictedlibrary, col_names=FALSE)
  predictedlib <- data.frame(predictedlib)
  
  #Count frequency of designed library sequences in Nanopore reads
  for (l in 1:nrow(predictedlib)) {
    predictedlib[l,2] = length(which(grepl(predictedlib[l,1], stack_seq)))}
  
  #Transform (to make compatible with column graph in Prism)
  predictedlib_frequency <- as.data.frame(t(predictedlib[,2]))
  colnames(predictedlib_frequency) <- predictedlib[,1]
  
  #Normalize frequency to total number of reads
  predictedlib_frequency = predictedlib_frequency/sum(df_frequency)

  #Add asterix (*) to sequences that are designed library variants
  for (m in 1:ncol(df_frequency)) {
    if (colnames(df_frequency)[m] %in% predictedlib[,1]) { #if sequences in Nanopore reads were intended
      colnames(df_frequency)[m] <- paste("*", colnames(df_frequency[m]), sep="")} #then add * to the sequence (colname)
  }
  
  df_frequency = df_frequency/sum(df_frequency) #normalize frequency to total number of reads
  df_i=paste("df_", i, sep = "")
  assign(df_i, df_frequency)
  list_i <- list(df_frequency) 
  mylist1 <- append(mylist1, list_i) #append to list
  
  list_i <- list(predictedlib_frequency) 
  mylist2 <- append(mylist2, list_i) #append to list
  
  #Editing efficiency
  efficiency_i = sum(predictedlib_frequency)*100
  efficiency_i = formatC(efficiency_i, digits = 3, format = "f")
  efficiency[i] <- efficiency_i
  
  #Count number of specific sequence in 1 defined region
  wildtypelist_i = length(which(grepl(wildtype[i], stack_seq)))/length(stack_seq)*100
  wildtypelist_i = formatC(wildtypelist_i, digits = 3, format = "f")
  wildtypelist[i] <- wildtypelist_i
}

#Formatting
names <- c()
for (i in 1:length(region)){
  names[i] <- paste(region[i], "-", as.character(as.numeric(region[i])+region_length[i]-1), sep = "") #define region
}
names(mylist1) <- names
names(mylist2) <- names

efficiency <- as.numeric(efficiency)
wildtypelist <- as.numeric(wildtypelist)
effwild <- data.frame(efficiency, wildtypelist)
colnames(effwild) <- c("Library", "Wildtype")
rownames(effwild) <- names


#Export to Prism, xlsx or csv files
write_pzfx(mylist1, paste(folder, filename, "-frequency-of-sequences", ".pzfx", sep = ""), row_names=TRUE) #export which sequences are present in the reads and their frequency
write.xlsx(mylist1, paste(folder, filename, "-frequency-of-sequences", ".xlsx", sep = ""), rowNames=FALSE) #use to write to xlsx file

write_pzfx(mylist2, paste(folder, filename, "-coverage-of-library-sequences", ".pzfx", sep = ""), row_names=TRUE) #export the fraction of designed library variants of total sequences
write.xlsx(mylist2, paste(folder, filename, "-coverage-of-library-sequences", ".xlsx", sep = ""), rowNames=FALSE) #use to write to xlsx file

write_pzfx(effwild, paste(folder, filename, "-editing-effiency", ".pzfx", sep = ""), row_names=TRUE) #export the fraction of designed library sequences and wildtype sequences
write.xlsx(effwild, paste(folder, filename, "-editing-effiency", ".xlsx", sep = ""), rowNames=TRUE) #use to write to xlsx file
#write.csv(effwild, paste(folder, filename, "-editing-effiency", ".csv", sep = ""), row.names=TRUE) #use to write to csv file









### PART 2: ANALYSIS PER PLASMID

#Extract the reads mapped from the first target region to the last target region
bamfile = paste(folder, filename, ".bam", sep = "")
indexfile = paste(folder, filename, sep = "") #indexfile is the .bai file
df1 <- c()
df2 <- c()
df3 <- c()

range = paste(indexfilename, ":", region[1], "-", as.character(as.numeric(region[length(region)])+region_length[length(region_length)]-1), sep = "") #define region

stack <- stackStringsFromBam(bamfile, indexfile, param=GRanges(range)) #extract sequences of target region
stack_seq = as.character(stack)

qual <- stackStringsFromBam(bamfile, indexfile, param=GRanges(range), what="qual") #extract per-base quality scores
qual_seq = as.character(qual)

for (i in (1:length(stack))) { #remove reads that do not span the full length from the first target region to the last target region
  if ((length(stack)+1-i) %in% which(grepl("+", stack_seq, fixed=TRUE))) {
    stack_seq <- stack_seq[-(length(stack)+1-i)]
    qual_seq <- qual_seq[-(length(stack)+1-i)]
  }
}


#For each target region, 
for (i in 1:length(region)){
  #import the designed library variants
  filename_predictedlibrary <- paste(folder, predictedlibrary[i], sep = "")
  predictedlib <- read_excel(filename_predictedlibrary, col_names=FALSE)
  predictedlib <- data.frame(predictedlib)
  
  #define start and end position of target region in stack_seq
  start=as.character(as.numeric(region[i])-(as.numeric(region[1])-1))
  end=as.character(as.numeric(start)+region_length[i]-1)
  
  #give labels to target sequences
  sequences <- c()
  for (j in (1:length(stack_seq))) {
    sequenceofinterest <- substr(stack_seq[j], start, end) #get RBS sequence of read i
    corrqual <- substr(qual_seq[j], start, end) #get per-base quality scores of the RBS sequence
    
    if (min(strtoi(charToRaw(corrqual),16L)) < qualityscore) {
      label = "Q" # Q = sequence contains at least 1 nt with per-base quality score lower than threshold
    }
    else if (sequenceofinterest %in% wildtype[i]) {
      label = "WT" # WT = sequence is identical to wildtype sequence
    }
    else if (sequenceofinterest %in% predictedlib[,1]) {
      label = "LIB" # LIB = sequence is a designed library variant
    }
    else {
      label = "OTHER" # OTHER = sequence is NOT wildtype or designed library variant
    }
    
    temp <- c(sequenceofinterest, corrqual, label)
    sequences = rbind(temp, sequences)
  }
  df1 <- cbind(df1, sequences[,3]) #labels
  df2 <- cbind(df2, sequences[,1]) #sequences
  df3 <- cbind(df3, sequences[,2]) #quality strings
}

#Remove reads that contain at least 1 nucleotide in any of the target sequences with per-base quality score lower than threshold
df_select <- c()
for (i in (1:length(stack_seq))) {
  if (!("Q" %in% df1[i,])) {
    df_select = rbind(df_select, df1[i,])
  }
}

#Look at df_select_seq to see which sequences are left
df_select_seq <- c()
for (i in (1:length(stack_seq))) {
  if (!("Q" %in% df1[i,])) {
    df_select_seq = rbind(df_select_seq, df2[i,])
  }
}

#Count how many library variants per read and add in additional column to df_select
vec <- vector(mode = "logical", length = nrow(df_select))
df_select <- cbind(df_select, new_col = vec)
for (i in (1:nrow(df_select))) {
  df_select[i,(length(region)+1)] = length(which(grepl("LIB", df_select[i,])))
}

#Count how many reads have 0, 1, 2, ... target regions modified and combine in 1 vector
result_count <- c()
for (i in (1:(length(region)+1))) {
  result_count[i] = length(which(grepl(i-1, df_select[,ncol(df_select)])))
}

result_fraction <- result_count/sum(result_count)
result_fraction = formatC(result_fraction, digits = 2, format = "f")

#Format for plotting
names <- as.character(0:length(region))
sample <- vector( "character" , length(region)+1)
result_count <- as.numeric(result_count)
result_fraction <- as.numeric(result_fraction)
result_count <- data.frame(sample, names, result_count)
result_fraction <- data.frame(sample, names, result_fraction)

#Plot
ggplot(result_count, aes(fill=names, y=result_count, x=sample)) + 
  geom_bar(position="stack", stat="identity") +
  ggtitle("Count") +
  xlab("") +
  ylab("Number of reads") + 
  scale_fill_discrete(name = "Number of mutated target loci per plasmid") +
  theme(axis.text.x=element_blank(),
        axis.ticks.x=element_blank())

ggplot(result_fraction, aes(fill=names, y=result_fraction, x=sample)) + 
  geom_bar(position="stack", stat="identity") +
  ggtitle("Fraction") +
  xlab("") +
  ylab("Fraction of reads") + 
  scale_fill_discrete(name = "Number of mutated target loci per plasmid") +
  theme(axis.text.x=element_blank(),
        axis.ticks.x=element_blank())

#Export to Prism, xlsx or csv file
result_fraction_plot <- as.data.frame(t(result_fraction[3]))
colnames(result_fraction_plot) <- as.character(result_fraction[,2])
output_pzfx_filename = paste(folder, filename, "-multiplex", ".pzfx", sep = "")
write_pzfx(result_fraction_plot, output_pzfx_filename, row_names=TRUE) #export fraction of reads that contain zero, one, two, three, ... modified target regions
write.xlsx(result_fraction_plot, paste(folder, filename, "-multiplex", ".xlsx", sep = ""), rowNames=TRUE) #use to write to xlsx file
#write.csv(result_fraction_plot, paste(folder, filename, "-multiplex", ".csv", sep = ""), row.names=TRUE) #use to write to csv file

#Results (data underlying graphs)
result_count
result_fraction


#Check number of reads used for analysis
print(numberofseq) #Number of reads used for PART 1: ANALYSIS PER LOCUS
print(nrow(df_select)) #Number of reads used for PART 2: ANALYSIS PER PLASMID

