#Install required packages: BiocManager, GenomicAlignments, ggplot2, pzfx

#Load required packages
library(GenomicAlignments)
library(ggplot2)
library(pzfx)
library(readxl)
library(openxlsx)

#Clear environment
rm(list = ls())

#Important! .bam and .bai files should have the same name and should have identical paths as this .R file



##RUN PART BELOW TO COUNT FRACTION OF SEQUENCES WITH 18-NT DELETION

#Essential information to be defined by user
folder = "/.../"
filename = "Galaxy150-26-6-B5" #bam and bai files should have same name and be in the same folder
region = "pUC19:721-738"
qualityscore = 50

#Import .bam and .bai files
bamfile = paste(folder, filename, ".bam", sep = "")
indexfile = paste(folder, filename, sep = "") #indexfile is .bai file

#Extract all sequences in the target region
stack <- stackStringsFromBam(bamfile, indexfile, param=GRanges(region)) # Extract the stacked reads from the position
stack_seq = as.character(stack)

#Extract per-base quality scores in the target region
qual <- stackStringsFromBam(bamfile, indexfile, param=GRanges(region), what="qual")
qual_seq = as.character(qual)

#Replace deletions in the quality scores (-) by another character, so that they are not discarded during quality control (deletion has quality score 45)
qual_seq <- chartr("-", "Z", qual_seq)

#Quality control: eliminate sequences with at least 1 nucleotide with a per-base quality score lower than the threshold
for (i in (1:length(qual_seq))) {
  if (min(strtoi(charToRaw(qual_seq[length(qual_seq)+1-i]),16L)) < qualityscore) {
      stack_seq = stack_seq[-(length(qual_seq)+1-i)]
   }
}

#Count number of deletions in the target region
indended_mutation = "------------------"
original_sequence = "AAAACCCTGGCGTTACCC"
occurrences_intended_mutation = which(grepl(indended_mutation, stack_seq))
occurrences_original_sequence = which(grepl(original_sequence, stack_seq))
frequency_intended_mutation = length(occurrences_intended_mutation)/length(stack_seq)*100
frequency_original_sequence = length(occurrences_original_sequence)/length(stack_seq)*100
print(frequency_intended_mutation) #Fraction of sequences with 18-nt deletion (%)
print(frequency_original_sequence) #Fraction of sequences with wildtype sequence (%)
print(length(stack_seq)) #Number of sequences used for analysis after quality control
print(100-frequency_intended_mutation-frequency_original_sequence) #Fraction of sequences not assigned to intended mutation or wildtype sequence (%)




##RUN PART BELOW TO COUNT FRACTION OF SEQUENCES WITH 18-NT INSERTION
#USE READS MAPPED TO SEQUENCE INCLUDING INSERTION (NOT WILDTYPE DNA SEQUENCE)
#THIS BECAUSE THIS SCRIPT CANNOT DETECT INSERTIONS, BUT CAN DETECT DELETIONS

#Essential information to be defined by user
folder = "/.../"
filename = "Galaxy157-26-6-B6-mappedto-pUC19-18INS" #bam and bai files should have same name and be in the same folder
region = "pUC19-18INS:1956-1973"
qualityscore = 50

#Import .bam and .bai files
bamfile = paste(folder, filename, ".bam", sep = "")
indexfile = paste(folder, filename, sep = "") #indexfile is .bai file

#Extract all sequences in the target region
stack <- stackStringsFromBam(bamfile, indexfile, param=GRanges(region)) # Extract the stacked reads from the position
stack_seq = as.character(stack)

#Extract per-base quality scores in the target region
qual <- stackStringsFromBam(bamfile, indexfile, param=GRanges(region), what="qual")
qual_seq = as.character(qual)

#Replace deletions in the quality scores (-) by another character, so that they are not discarded during quality control (deletion has quality score 45)
qual_seq <- chartr("-", "Z", qual_seq)

#Quality control: eliminate sequences with at least 1 nucleotide with a per-base quality score lower than the threshold
for (i in (1:length(qual_seq))) {
  if (min(strtoi(charToRaw(qual_seq[length(qual_seq)+1-i]),16L)) < qualityscore) {
    stack_seq = stack_seq[-(length(qual_seq)+1-i)]
  }
}

#Count number of deletions in the target region
indended_mutation = "AAAGTATCAGGACGGTAT"
original_sequence = "------------------"
occurrences_intended_mutation = which(grepl(indended_mutation, stack_seq))
occurrences_original_sequence = which(grepl(original_sequence, stack_seq))
frequency_intended_mutation = length(occurrences_intended_mutation)/length(stack_seq)*100
frequency_original_sequence = length(occurrences_original_sequence)/length(stack_seq)*100
print(frequency_intended_mutation) #Fraction of sequences with 18-nt insertion (%)
print(frequency_original_sequence) #Fraction of sequences with wildtype sequence (%)
print(length(stack_seq)) #Number of sequences used for analysis after quality control
print(100-frequency_intended_mutation-frequency_original_sequence) #Fraction of sequences not assigned to intended mutation or wildtype sequence (%)
 



##RUN PART BELOW TO COUNT FRACTION OF SEQUENCES WITH 18-NT SUBSTITUTION

#Essential information to be defined by user
folder = "/.../"
filename = "Galaxy275-13-07-C2-mappedto-pUC19-penalties32-72" #bam and bai files should have same name and be in the same folder
region = "pUC19:722-739"
qualityscore = 50

#Import .bam and .bai files
bamfile = paste(folder, filename, ".bam", sep = "")
indexfile = paste(folder, filename, sep = "") #indexfile is .bai file

#Extract all sequences in the target region
stack <- stackStringsFromBam(bamfile, indexfile, param=GRanges(region)) # Extract the stacked reads from the position
stack_seq = as.character(stack)

#Extract per-base quality scores in the target region
qual <- stackStringsFromBam(bamfile, indexfile, param=GRanges(region), what="qual")
qual_seq = as.character(qual)

#Quality control: eliminate sequences with at least 1 nucleotide with a per-base quality score lower than the threshold
for (i in (1:length(qual_seq))) {
  if (min(strtoi(charToRaw(qual_seq[length(qual_seq)+1-i]),16L)) < qualityscore) {
    stack_seq = stack_seq[-(length(qual_seq)+1-i)]
  }
}

#Count number of modified sequences in the target region
indended_mutation = "TCGATTCCTTCAGCGGAC"
original_sequence = "AAACCCTGGCGTTACCCA"
occurrences_intended_mutation = which(grepl(indended_mutation, stack_seq))
occurrences_original_sequence = which(grepl(original_sequence, stack_seq))
frequency_intended_mutation = length(occurrences_intended_mutation)/length(stack_seq)*100
frequency_original_sequence = length(occurrences_original_sequence)/length(stack_seq)*100

## Print results
print(frequency_intended_mutation) #Fraction of sequences with 18-nt substitution (%)
print(frequency_original_sequence) #Fraction of sequences with wildtype sequence (%)
print(length(stack_seq)) #Number of sequences used for analysis after quality control
print(100-frequency_intended_mutation-frequency_original_sequence) #Fraction of sequences not assigned to intended mutation or wildtype sequence (%)


